package dev.fyloz.memorycache

import dev.fyloz.memorycache.utils.removeFirst
import java.time.Instant
import java.util.*

/**
 * A key-value memory cache allowing for expiring keys.
 *
 * The key's expiration can be setup with two methods:
 *  - After a number of access to the cache
 *  - After a duration
 *
 *  Either methods can be used, or both. But at least one need to be configured.
 *  Note that using a single method is more optimized, since the keys can be sorted easily.
 *
 *  For both methods, the lifetime of the key will be renewed after it has been set or accessed.
 *
 *  @param K The type of the keys.
 *  @param V The type of the values.
 *  @property maxAccessCount The number of access to the cache before a key expires.
 *  @property maxLifetimeSeconds The duration of a key before its expiration in seconds.
 */
class ExpiringMemoryCache<K, V>(
    private val maxAccessCount: Long = -1,
    private val maxLifetimeSeconds: Long = -1
) : BaseMemoryCache<K, V>() {
    private val keys: TreeSet<ExpiringKey<K>>
    private var accessCount = 0L

    private val maxLifetimeMillis = maxLifetimeSeconds * 1000
    private val expireByAccessCount: Boolean = maxAccessCount > 0
    private val expireByAccessTime: Boolean = maxLifetimeSeconds > 0

    init {
        // Determine the correct key comparator to use
        val keysComparator: Comparator<ExpiringKey<*>> = if (expireByAccessCount) {
            ExpiringKey.accessCountComparator
        } else if (expireByAccessTime) {
            ExpiringKey.timeComparator
        } else {
            throw InvalidExpiringCacheOptionsException()
        }

        keys = TreeSet(keysComparator)
    }

    override fun get(key: K): V? {
        accessCount++

        cleanup()
        renewKey(key)

        return super.get(key)
    }

    override fun set(key: K, value: V) {
        super.set(key, value)

        renewKey(key)
    }

    override fun remove(key: K) {
        super.remove(key)
        keys.removeFirst { it.key == key }
    }

    override fun clear() {
        super.clear()
        keys.clear()

        accessCount = 0
    }

    private fun renewKey(key: K) {
        keys.removeFirst { it.key == key }
        keys.add(getNewExpiringKey(key))
    }

    private fun cleanup() {
        with(keys.iterator()) {
            while (hasNext()) {
                val expiringKey = next()
                if (expireByAccessCount) {
                    if (accessCount - expiringKey.lastAccessCount > maxAccessCount) {
                        removeExpiringKey(expiringKey)
                    } else if (!expireByAccessTime) {
                        break   // Keys are sorted, so if this key is not expired, the next ones are not
                    }
                }
                if (expireByAccessTime) {
                    if (getCurrentTime() - expiringKey.lastAccessMillis > maxLifetimeMillis) {
                        removeExpiringKey(expiringKey)
                    } else if (!expireByAccessCount) {
                        break
                    }
                }
            }
        }
    }

    private fun MutableIterator<ExpiringKey<K>>.removeExpiringKey(expiringKey: ExpiringKey<K>) {
        remove()
        super.remove(expiringKey.key)
    }

    private fun getNewExpiringKey(key: K) =
        ExpiringKey(key, accessCount, getCurrentTime())

    private fun getCurrentTime() =
        Instant.now().toEpochMilli()
}

private data class ExpiringKey<K>(val key: K, var lastAccessCount: Long, var lastAccessMillis: Long) {
    companion object {
        internal val accessCountComparator: Comparator<ExpiringKey<*>> =
            Comparator.comparingLong<ExpiringKey<*>> { it.lastAccessCount }
                .thenComparingLong { it.lastAccessMillis }
                .thenComparingInt { it.key.hashCode() }

        internal val timeComparator: Comparator<ExpiringKey<*>> =
            Comparator.comparingLong<ExpiringKey<*>> { it.lastAccessMillis }
                .thenComparingLong { it.lastAccessCount }
                .thenComparingInt { it.key.hashCode() }
    }
}

internal class InvalidExpiringCacheOptionsException :
    RuntimeException("An expiration cache must have at least one expiration method configured")