package dev.fyloz.memorycache

import io.kotest.assertions.throwables.shouldThrow
import io.kotest.core.spec.style.ShouldSpec
import io.kotest.matchers.shouldBe
import kotlinx.coroutines.Dispatchers
import kotlinx.coroutines.withContext

class ExpiringMemoryCacheTest : ShouldSpec({
    context("expiration by access count") {
        val maxAccessCount = 5L
        val cache = ExpiringMemoryCache<Int, String>(maxAccessCount = maxAccessCount)

        val accessedKey = 0
        val expiringKey = Int.MAX_VALUE

        should("removes expired keys") {
            // Arrange
            cache[accessedKey] = accessedKey.toString()
            cache[expiringKey] = expiringKey.toString()

            // Act
            for (i in 0..maxAccessCount) {
                cache[accessedKey]
            }

            // Assert
            val containsUnusedKey = expiringKey in cache
            containsUnusedKey shouldBe false
        }

        should("renew keys on access") {
            // Arrange
            cache[accessedKey] = accessedKey.toString()
            cache[expiringKey] = expiringKey.toString()

            for (i in 0 until maxAccessCount - 1) {
                cache[accessedKey]
            }

            // Act
            cache[expiringKey]

            // Assert
            val containsUnusedKey = expiringKey in cache
            containsUnusedKey shouldBe true
        }
    }

    context("expiration by access time") {
        val maxLifetimeSeconds = 2L
        val cache = ExpiringMemoryCache<Int, String>(maxLifetimeSeconds = maxLifetimeSeconds)

        val key = Int.MAX_VALUE

        suspend fun wait(seconds: Long) =
            withContext(Dispatchers.IO) {
                Thread.sleep(seconds * 1000)
            }

        should("removes expired keys") {
            // Arrange
            cache[key] = key.toString()

            // Act
            wait(maxLifetimeSeconds)
            cache[key]

            // Assert
            val containsKey = key in cache
            containsKey shouldBe false
        }

        should("renew keys on access") {
            // Arrange
            cache[key] = key.toString()

            // Act
            wait(maxLifetimeSeconds / 2)
            cache[key]  // Access key before expiration
            wait(maxLifetimeSeconds / 2)
            cache[key]

            // Assert
            val containsKey = key in cache
            containsKey shouldBe true
        }
    }

    context("no expiration method defined") {
        should("throws InvalidExpiringCacheOptionsException") {
            // Arrange
            // Act
            // Assert
            shouldThrow<InvalidExpiringCacheOptionsException> { ExpiringMemoryCache<Int, String>() }
        }
    }
})